"""run 'pytest' from the '.../Code/' directory"""
import pytest
from test.utils import (
    check_model_status,
    default_max_time,
)

from src.model_checker.syntactic import (
    OperatorCollection,
)

from src.model_checker.semantic import (
    Proposition,
    Semantics,
)

from src.model_checker.defined import (
    NegationOperator, # extensional
    AndOperator,
    OrOperator,
    ConditionalOperator, # extensional defined 
    BiconditionalOperator,
    TopOperator, # extremal
    BotOperator,
    CounterfactualOperator, # counterfactual
    NecessityOperator, # modal
    PossibilityOperator,
    DefNecessityOperator,
    DefPossibilityOperator,  # modal defined
)

semantics = Semantics
proposition = Proposition
operators = OperatorCollection(
    NegationOperator, # extensional
    AndOperator,
    OrOperator,
    ConditionalOperator, # extensional defined 
    BiconditionalOperator,
    TopOperator, # extremal
    BotOperator,
    CounterfactualOperator, # counterfactual
    NecessityOperator, # modal
    PossibilityOperator,
    DefNecessityOperator,
    DefPossibilityOperator,  # modal defined
)

max_time = default_max_time

###############################
##### MODAL COUNTERMODELS #####
###############################

@pytest.mark.timeout(max_time)
def test_ML_CM1():
    """ NECESSITATED ARGUMENTS COUNTERFACTUAL MODUS PONENS """
    premises = ['\\Box A','(A \\rightarrow B)']
    conclusions = ['\\Box B']
    settings = {
        'N' : 3,
        'desired_status' : True,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML_CM2():
    """ COUNTERFACTUAL IMPLIES STRICT CONDITIONAL """
    premises = ['(A \\boxright B)']
    conclusions = ['\\Box (A \\rightarrow B)']
    settings = {
        'N' : 3,
        'desired_status' : True,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )






################################
######### MODAL LOGIC ##########
################################

@pytest.mark.timeout(max_time)
def test_ML1():
    premises = ['\\Box (A \\rightarrow B)']
    conclusions = ['(A \\boxright B)']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML2():
    # K AXIOM (BOX)
    premises = ['\\Box (A \\rightarrow B)']
    conclusions = ['(\\Box A \\rightarrow \\Box B)']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML3():
    # K AXIOM (TOP)
    premises = ['(\\top \\boxright (A \\rightarrow B))']
    conclusions = ['((\\top \\boxright A) \\rightarrow (\\top \\boxright B))']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML4():
    # T AXIOM (TOP)
    premises = ['(\\top \\boxright A)']
    conclusions = ['A']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML5():
    # T AXIOM (BOX)
    premises = ['\\Box A']
    conclusions = ['A']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML6():
    # 4 AXIOM (TOP)
    premises = ['(\\top \\boxright A)']
    conclusions = ['(\\top \\boxright (\\top \\boxright A))']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML7():
    # 4 AXIOM (BOX)
    premises = ['\\Box A']
    conclusions = ['\\Box \\Box A']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML8():
    # B AXIOM (TOP)
    # NOTE: with Z3 quantifiers MIT ran for 1600 seconds; now .0328 seconds locally
    premises = ['A']
    conclusions = ['(\\top \\boxright \\neg (\\top \\boxright \\neg A))']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML9():
    # B AXIOM (BOX)
    premises = ['A']
    conclusions = ['\\Box \\Diamond A']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML10():
    # 5 AXIOM (TOP)
    # SLOW: 12.9 seconds locally
    premises = ['(\\top \\boxright A)']
    conclusions = ['(\\top \\boxright \\neg (\\top \\boxright \\neg A))']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML11():
    # 5 AXIOM (BOX)
    premises = ['\\Box A']
    conclusions = ['\\Box \\Diamond A']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML12():
    # BOX-TO-\\top EQUIVALENCE
    premises = ['\\Box A']
    conclusions = ['(\\top \\boxright A)']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML13():
    # # TOP-TO-BOX EQUIVALENCE
    premises = ['(\\top \\boxright A)']
    conclusions = ['\\Box A']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_ML14():
    # NECESSARY EQUIVALENCE
    premises = []
    conclusions = ['\\Box ((A \\vee \\neg A) \\leftrightarrow (B \\vee \\neg B))']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )



"""From the ../Code/ directory, run: python -m src.model_checker.examples"""
from src.model_checker.builder import (
    make_model_for,
    find_max_N,
    run_comparison,
    save_comparisons,
)
 
from src.model_checker.semantic import (
    Semantics,
    ImpositionSemantics,
    Proposition,
)

from src.model_checker.defined import (
    ConditionalOperator, BiconditionalOperator, # extensional defined
    DefEssenceOperator, DefGroundOperator, # constitutive defined
    MightCounterfactualOperator, # counterfactual
    MightImpositionOperator,
)

from src.model_checker.primitive import (
    AndOperator, NegationOperator, OrOperator, # extensional
    TopOperator, BotOperator, # top and bottom zero-place operators
    IdentityOperator, GroundOperator, EssenceOperator, # constitutive
    NecessityOperator, PossibilityOperator, # modal
    CounterfactualOperator, # counterfactual
    ImpositionOperator, # counterfactual
)

from src.model_checker import syntactic

######################
### SETUP LANGUAGE ###
######################

operators = syntactic.OperatorCollection(
    AndOperator, NegationOperator, OrOperator, # extensional
    ConditionalOperator, BiconditionalOperator, # extensional defined
    TopOperator, BotOperator, # top and bottom zero-place operators
    IdentityOperator, GroundOperator, EssenceOperator, # constitutive
    DefEssenceOperator, DefGroundOperator, # constitutive defined
    NecessityOperator, PossibilityOperator, # modal
    CounterfactualOperator, MightCounterfactualOperator, # counterfactual
    ImpositionOperator, MightImpositionOperator, # Fine
)


################
### DEFAULTS ###
################

settings = {
    'N' : 3,
    'contingent' : True,
    'non_empty' : True,
    'non_null' : True,
    'disjoint' : False,
    'print_impossible' : True,
    'max_time' : 1,
}


############################
##### REMAINING ISSUES #####
############################

# # DOES NOT FIND MODEL
# # THIS WAS EXTRA HARD BEFORE ALSO
# N = 4
# premises = ['(A \\boxright (B \\boxright C))']
# conclusions = ['((A \\wedge B) \\boxright C)']




#####################
### COUNTERMODELS ###
#####################

# # CF_CM1: COUNTERFACTUAL ANTECEDENT STRENGTHENING
CF_CM1_premises = ['(A \\boxright C)']
CF_CM1_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM1_example = [
    CF_CM1_premises,
    CF_CM1_conclusions,
]

# # CF_CM2: MIGHT COUNTERFACTUAL ANTECEDENT STRENGTHENING
CF_CM2_premises = ['(A \\circleright C)']
CF_CM2_conclusions = ['((A \\wedge B) \\circleright C)']
CF_CM2_example = [
    CF_CM2_premises,
    CF_CM2_conclusions,
]

# # CF_CM3: COUNTERFACTUAL ANTECEDENT STRENGTHENING WITH POSSIBILITY
CF_CM3_premises = ['(A \\boxright C)', '\\Diamond (A \\wedge B)']
CF_CM3_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM3_example = [
    CF_CM3_premises,
    CF_CM3_conclusions,
]

# # CF_CM4: COUNTERFACTUAL ANTECEDENT STRENGTHENING WITH NEGATION
# N = 4
CF_CM4_premises = ['\\neg A','(A \\boxright C)']
CF_CM4_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM4_example = [
    CF_CM4_premises,
    CF_CM4_conclusions,
]

# # CF_CM5: COUNTERFACTUAL DOUBLE ANTECEDENT STRENGTHENING
# N = 4
CF_CM5_premises = ['(A \\boxright C)','(B \\boxright C)']
CF_CM5_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM5_example = [
    CF_CM5_premises,
    CF_CM5_conclusions,
]

# # CF_CM6: WEAKENED MONOTONICITY
# N = 3
CF_CM6_premises = ['(A \\boxright B)','(A \\boxright C)']
CF_CM6_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM6_example = [
    CF_CM6_premises,
    CF_CM6_conclusions,
]
# settings['contingent'] = False

# # CF_CM7: COUNTERFACTUAL CONTRAPOSITION
# N = 3
CF_CM7_premises = ['(A \\boxright B)']
CF_CM7_conclusions = ['(\\neg B \\boxright \\neg A)']
CF_CM7_example = [
    CF_CM7_premises,
    CF_CM7_conclusions,
]

# # CF_CM8: COUNTERFACTUAL CONTRAPOSITION WITH NEGATION
# N = 4
CF_CM8_premises = ['\\neg B','(A \\boxright B)']
CF_CM8_conclusions = ['(\\neg B \\boxright \\neg A)']
CF_CM8_example = [
    CF_CM8_premises,
    CF_CM8_conclusions,
]

# # CF_CM9: COUNTERFACTUAL CONTRAPOSITION WITH TWO NEGATIONS
# N = 4
CF_CM9_premises = ['\\neg A','\\neg B','(A \\boxright B)']
CF_CM9_conclusions = ['(\\neg B \\boxright \\neg A)']
CF_CM9_example = [
    CF_CM9_premises,
    CF_CM9_conclusions,
]

# # CF_CM10: TRANSITIVITY
# N = 3
CF_CM10_premises = ['(A \\boxright B)','(B \\boxright C)']
CF_CM10_conclusions = ['(A \\boxright C)']
CF_CM10_example = [
    CF_CM10_premises,
    CF_CM10_conclusions,
]

# # CF_CM11: COUNTERFACTUAL TRANSITIVITY WITH NEGATION
# N = 3
CF_CM11_premises = ['\\neg A','(A \\boxright B)','(B \\boxright C)']
CF_CM11_conclusions = ['(A \\boxright C)']
CF_CM11_example = [
    CF_CM11_premises,
    CF_CM11_conclusions,
]

# # CF_CM12: COUNTERFACTUAL TRANSITIVITY WITH TWO NEGATIONS
# N = 4
CF_CM12_premises = ['\\neg A','\\neg B','(A \\boxright B)','(B \\boxright C)']
CF_CM12_conclusions = ['(A \\boxright C)']
CF_CM12_example = [
    CF_CM12_premises,
    CF_CM12_conclusions,
]

# # CF_CM13: SOBEL SEQUENCE
# N = 3
CF_CM13_premises = [
    '(A \\boxright X)',
    '\\neg ((A \\wedge B) \\boxright X)',
    '(((A \\wedge B) \\wedge C) \\boxright X)',
    '\\neg ((((A \\wedge B) \\wedge C) \\wedge D) \\boxright X)',
    '(((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\boxright X)',
    '\\neg ((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\boxright X)',
    '(((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\wedge G) \\boxright X)', # 327.2 seconds on the MIT servers; now .01244 seconds
]
CF_CM13_conclusions = []
CF_CM13_example = [
    CF_CM13_premises,
    CF_CM13_conclusions,
]

# # CF_CM14: SOBEL SEQUENCE WITH POSSIBILITY (N = 3)
# N = 3
CF_CM14_premises = [
    '\\Diamond A',
    '(A \\boxright X)',
    '\\Diamond (A \\wedge B)',
    '\\neg ((A \\wedge B) \\boxright X)', # N = 4: 155.4 seconds on the MIT servers; .1587 seconds in old version; and now .0122 seconds
    '\\Diamond ((A \\wedge B) \\wedge C)',
    '(((A \\wedge B) \\wedge C) \\boxright X)',
    '\\Diamond (((A \\wedge B) \\wedge C) \\wedge D)',
    '\\neg ((((A \\wedge B) \\wedge C) \\wedge D) \\boxright X)',
    '\\Diamond ((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E)',
    '(((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\boxright X)', # ? seconds
    '\\Diamond (((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F)',
    '\\neg ((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\boxright X)', # ? seconds
    '\\Diamond ((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\wedge G)',
    '(((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\wedge G) \\boxright X)', # ? seconds
]
CF_CM14_conclusions = []
CF_CM14_example = [
    CF_CM14_premises,
    CF_CM14_conclusions,
]

# # CF_CM15: COUNTERFACTUAL EXCLUDED MIDDLE
# N = 3
CF_CM15_premises = ['\\neg A']
CF_CM15_conclusions = ['(A \\boxright B)','(A \\boxright \\neg B)']
CF_CM15_example = [
    CF_CM15_premises,
    CF_CM15_conclusions,
]

# # CF_CM16: SIMPLIFICATION OF DISJUNCTIVE CONSEQUENT
# N = 3
CF_CM16_premises = ['\\neg A','(A \\boxright (B \\vee C))']
CF_CM16_conclusions = ['(A \\boxright B)','(A \\boxright C)']
CF_CM16_example = [
    CF_CM16_premises,
    CF_CM16_conclusions,
]

# # CF_CM17: INTRODUCTION OF DISJUNCTIVE ANTECEDENT
# N = 4
CF_CM17_premises = ['(A \\boxright C)','(B \\boxright C)']
CF_CM17_conclusions = ['((A \\vee B) \\boxright C)']
CF_CM17_example = [
    CF_CM17_premises,
    CF_CM17_conclusions,
]

# # CF_CM18: MUST FACTIVITY
# N = 3
CF_CM18_premises = ['A','B']
CF_CM18_conclusions = ['(A \\boxright B)']
CF_CM18_example = [
    CF_CM18_premises,
    CF_CM18_conclusions,
]

# # CF_CM19: COUNTERFACTUAL EXPORTATION
# N = 3
CF_CM19_premises = ['((A \\wedge B) \\boxright C)']
CF_CM19_conclusions = ['(A \\boxright (B \\boxright C))']
CF_CM19_example = [
    CF_CM19_premises,
    CF_CM19_conclusions,
]

# # CF_CM20: COUNTERFACTUAL EXPORTATION WITH POSSIBILITY
# N = 3
CF_CM20_premises = ['((A \\wedge B) \\boxright C)','\\Diamond (A \\wedge B)']
CF_CM20_conclusions = ['(A \\boxright (B \\boxright C))']
CF_CM20_example = [
    CF_CM20_premises,
    CF_CM20_conclusions,
]

# # CF_CM21:
# N = 3
CF_CM21_premises = ['\\neg A','\\neg (A \\boxright B)']
CF_CM21_conclusions = ['(A \\boxright \\neg B)']
CF_CM21_example = [
    CF_CM21_premises,
    CF_CM21_conclusions
]




############################
### LOGICAL CONSEQUENCES ###
############################

# # CF_T1: COUNTERFACTUAL IDENTITY
# N = 3
CF_T1_premises = []
CF_T1_conclusions = ['(A \\boxright A)']
CF_T1_example = [
    CF_T1_premises,
    CF_T1_conclusions
]

# # CF_T2: COUNTERFACTUAL MODUS PONENS
# N = 3
CF_T2_premises = ['A','(A \\boxright B)']
CF_T2_conclusions = ['B']
CF_T2_example = [
    CF_T2_premises,
    CF_T2_conclusions
]

# # CF_T3: WEAKENED TRANSITIVITY
# N = 3
CF_T3_premises = ['(A \\boxright B)','((A \\wedge B) \\boxright C)']
CF_T3_conclusions = ['(A \\boxright C)']
CF_T3_example = [
    CF_T3_premises,
    CF_T3_conclusions
]

# # CF_T4: ANTECEDENT DISJUNCTION TO CONJUNCTION
# N = 3
CF_T4_premises = ['((A \\vee B) \\boxright C)']
CF_T4_conclusions = ['((A \\wedge B) \\boxright C)']
CF_T4_example = [
    CF_T4_premises,
    CF_T4_conclusions
]

# # CF_T5: SIMPLIFICATION OF DISJUNCTIVE ANTECEDENT
# N = 4
CF_T5_premises = ['((A \\vee B) \\boxright C)']
CF_T5_conclusions = ['(A \\boxright C)']
CF_T5_example = [
    CF_T5_premises,
    CF_T5_conclusions
]

# # CF_T6: DOUBLE SIMPLIFICATION OF DISJUNCTIVE ANTECEDENT
# N = 3
CF_T6_premises = ['((A \\vee B) \\boxright C)']
CF_T6_conclusions = ['((A \\boxright C) \\wedge (B \\boxright C))']
CF_T6_example = [
    CF_T6_premises,
    CF_T6_conclusions
]

# # CF_T7:
# N = 3
CF_T7_premises = [
    '(A \\boxright C)',
    '(B \\boxright C)',
    '((A \\wedge B) \\boxright C)',
]
CF_T7_conclusions = ['((A \\vee B) \\boxright C)']
CF_T7_example = [
    CF_T7_premises,
    CF_T7_conclusions
]


# # CF_T8:
# N = 3
CF_T8_premises = ['(A \\boxright (B \\wedge C))']
CF_T8_conclusions = ['(A \\boxright B)']
CF_T8_example = [
    CF_T8_premises,
    CF_T8_conclusions
]

# # CF_T9:
# N = 3
CF_T9_premises = ['(A \\boxright B)','(A \\boxright C)']
CF_T9_conclusions = ['(A \\boxright (B \\wedge C))']
CF_T9_example = [
    CF_T9_premises,
    CF_T9_conclusions
]

# # # CF_T_T10: FACTIVITY MIGHT
# N = 4
CF_T10_premises = ['A','B']
CF_T10_conclusions = ['(A \\circleright B)']
CF_T10_example = [
    CF_T10_premises,
    CF_T10_conclusions
]

# # # CF_T_T11: DEFINITION OF NEC
# N = 4
CF_T11_premises = ['(\\neg A \\boxright \\bot)']
CF_T11_conclusions = ['(\\top \\boxright A)']
CF_T11_example = [
    CF_T11_premises,
    CF_T11_conclusions,
]




#########################################
### GENERATE Z3 CONSTRAINTS AND PRINT ###
#########################################

### NOTE: run below for individual tests

premises, conclusions = CF_CM2_example
# premises, conclusions = CF_T9_example
model_structure = make_model_for(
    settings,
    premises,
    conclusions,
    Semantics,
    Proposition,
    operators,
)
model_structure.print_all()

# premises, conclusions = CF_T10_example
# max_N = find_max_N(
#     settings,
#     premises,
#     conclusions,
#     Semantics,
#     Proposition,
#     operators,
# )

CF_examples = {
    "CF_T1" : CF_T1_example,
    "CF_T2" : CF_T2_example,
    "CF_T3" : CF_T3_example,
    "CF_T4" : CF_T4_example,
    "CF_T5" : CF_T5_example,
    "CF_T6" : CF_T6_example,
    "CF_T7" : CF_T7_example,
    "CF_T8" : CF_T8_example,
    "CF_T9" : CF_T9_example,
    "CF_T10" : CF_T10_example,
    "CF_T11" : CF_T11_example,
}

CM_examples = {
    "CF_CM1" : CF_CM1_example,
    "CF_CM2" : CF_CM2_example,
    "CF_CM3" : CF_CM3_example,
    "CF_CM4" : CF_CM4_example,
    "CF_CM5" : CF_CM5_example,
    "CF_CM6" : CF_CM6_example,
    "CF_CM7" : CF_CM7_example,
    "CF_CM8" : CF_CM8_example,
    "CF_CM9" : CF_CM9_example,
    "CF_CM10" : CF_CM10_example,
    "CF_CM11" : CF_CM11_example,
    "CF_CM12" : CF_CM12_example,
    "CF_CM13" : CF_CM13_example,
    "CF_CM14" : CF_CM14_example,
    "CF_CM15" : CF_CM15_example,
    "CF_CM16" : CF_CM16_example,
    "CF_CM17" : CF_CM17_example,
    "CF_CM18" : CF_CM18_example,
    "CF_CM19" : CF_CM19_example,
    "CF_CM20" : CF_CM20_example,
    "CF_CM21" : CF_CM21_example,
}

default_theory = [
    Semantics,
    Proposition,
    operators,
]

imposition_theory = [
    ImpositionSemantics,
    Proposition,
    operators,
]

# # Run comparison
# run_comparison(default_theory, imposition_theory, settings, CF_examples)
# run_comparison(default_theory, imposition_theory, settings, CM_examples)

# # Store output in a file
# save_comparisons(default_theory, imposition_theory, settings, CF_examples)
# save_comparisons(default_theory, imposition_theory, settings, CM_examples)

"""

There is a 3-model of:

Premise:
1. (A \circleright C)

Conclusion:
2. ((A \wedge B) \circleright C)

State Space:
  #b000 = □
  #b001 = a (world)
  #b010 = b
  #b011 = a.b (impossible)
  #b100 = c
  #b101 = a.c (impossible)
  #b110 = b.c (world)
  #b111 = a.b.c (impossible)

The evaluation world is: [34mb.c[0m

INTERPRETED PREMISE:

1.  [32m|(A \circleright C)| = < {□}, ∅ >[0m  [32m(True in b.c)[0m
      [37m|A| = < {a}, {b.c} >[0m  [33m(False in b.c)[0m
      [36m|A|-alternatives to b.c = {a}[0m
        [37m|C| = < {a}, {b.c, c} >[0m  [33m(True in a)[0m

INTERPRETED CONCLUSION:

2.  [31m|((A \wedge B) \circleright C)| = < ∅, {□} >[0m  [31m(False in b.c)[0m
      [37m|(A \wedge B)| = < {a.b.c}, {a, a.b.c, b.c} >[0m  [33m(False in b.c)[0m
        [37m|A| = < {a}, {b.c} >[0m  [33m(False in b.c)[0m
        [37m|B| = < {b.c}, {a} >[0m  [33m(True in b.c)[0m
      [36m|(A \wedge B)|-alternatives to b.c = ∅[0m

Z3 run time: 0.0084 seconds

Total run time: 1.103 seconds

Run time: 0.0084 seconds

"""

"""

There is a 3-model of:

Premise:
1. (A \circleright C)

Conclusion:
2. ((A \wedge B) \circleright C)

State Space:
  #b000 = □
  #b001 = a (world)
  #b010 = b
  #b011 = a.b (impossible)
  #b100 = c
  #b101 = a.c (impossible)
  #b110 = b.c (world)
  #b111 = a.b.c (impossible)

The evaluation world is: [34mb.c[0m

INTERPRETED PREMISE:

1.  [32m|(A \circleright C)| = < {□}, ∅ >[0m  [32m(True in b.c)[0m
      [37m|A| = < {a}, {b.c} >[0m  [33m(False in b.c)[0m
      [36m|A|-alternatives to b.c = {a}[0m
        [37m|C| = < {a}, {b.c, c} >[0m  [33m(True in a)[0m

INTERPRETED CONCLUSION:

2.  [31m|((A \wedge B) \circleright C)| = < ∅, {□} >[0m  [31m(False in b.c)[0m
      [37m|(A \wedge B)| = < {a.b.c}, {a, a.b.c, b.c} >[0m  [33m(False in b.c)[0m
        [37m|A| = < {a}, {b.c} >[0m  [33m(False in b.c)[0m
        [37m|B| = < {b.c}, {a} >[0m  [33m(True in b.c)[0m
      [36m|(A \wedge B)|-alternatives to b.c = ∅[0m

Z3 run time: 0.0094 seconds

Total run time: 11.8988 seconds

Run time: 0.0094 seconds

"""

"""

There is a 3-model of:

Premise:
1. (A \circleright C)

Conclusion:
2. ((A \wedge B) \circleright C)

State Space:
  #b000 = □
  #b001 = a (world)
  #b010 = b
  #b011 = a.b (impossible)
  #b100 = c
  #b101 = a.c (impossible)
  #b110 = b.c (world)
  #b111 = a.b.c (impossible)

The evaluation world is: [34mb.c[0m

INTERPRETED PREMISE:

1.  [32m|(A \circleright C)| = < {□}, ∅ >[0m  [32m(True in b.c)[0m
      [37m|A| = < {a}, {b.c} >[0m  [33m(False in b.c)[0m
      [36m|A|-alternatives to b.c = {a}[0m
        [37m|C| = < {a}, {b.c, c} >[0m  [33m(True in a)[0m

INTERPRETED CONCLUSION:

2.  [31m|((A \wedge B) \circleright C)| = < ∅, {□} >[0m  [31m(False in b.c)[0m
      [37m|(A \wedge B)| = < {a.b.c}, {a, a.b.c, b.c} >[0m  [33m(False in b.c)[0m
        [37m|A| = < {a}, {b.c} >[0m  [33m(False in b.c)[0m
        [37m|B| = < {b.c}, {a} >[0m  [33m(True in b.c)[0m
      [36m|(A \wedge B)|-alternatives to b.c = ∅[0m

Z3 run time: 0.0086 seconds

Total run time: 4.7602 seconds

Run time: 0.0086 seconds

"""

"""

There is a 3-model of:

Premise:
1. (A \circleright C)

Conclusion:
2. ((A \wedge B) \circleright C)

State Space:
  #b000 = □
  #b001 = a (world)
  #b010 = b
  #b011 = a.b (impossible)
  #b100 = c
  #b101 = a.c (impossible)
  #b110 = b.c (world)
  #b111 = a.b.c (impossible)

The evaluation world is: b.c

INTERPRETED PREMISE:

1.USE COLORS: False
  |(A \circleright C)| = < {□}, ∅ >  (True in b.c)
USE COLORS: False
      |A| = < {a}, {b.c} >  (False in b.c)
      [36m|A|-alternatives to b.c = {a}[0m
USE COLORS: False
        |C| = < {a}, {b.c, c} >  (True in a)

INTERPRETED CONCLUSION:

2.USE COLORS: False
  |((A \wedge B) \circleright C)| = < ∅, {□} >  (False in b.c)
USE COLORS: False
      |(A \wedge B)| = < {a.b.c}, {a, a.b.c, b.c} >  (False in b.c)
USE COLORS: False
        |A| = < {a}, {b.c} >  (False in b.c)
USE COLORS: False
        |B| = < {b.c}, {a} >  (True in b.c)
      [36m|(A \wedge B)|-alternatives to b.c = ∅[0m

Z3 run time: 0.0083 seconds

Total run time: 2.7498 seconds

Run time: 0.0083 seconds

"""



from . import syntactic

from .primitive import (
    AndOperator, NegationOperator, OrOperator, # extensional
    TopOperator, BotOperator, # top and bottom zero-place operators
    IdentityOperator, GroundOperator, EssenceOperator, # constitutive
    RelevanceOperator, # relevance
    CounterfactualOperator, # counterfactual
    ImpositionOperator,
    NecessityOperator, PossibilityOperator, # modal
)

##############################################################################
######################## DEFINED EXTENSIONAL OPERATORS #######################
##############################################################################

class ConditionalOperator(syntactic.DefinedOperator):

    name = "\\rightarrow"
    arity = 2

    def derived_definition(self, leftarg, rightarg):
        return [OrOperator, [NegationOperator, leftarg], rightarg]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Prints the proposition for sentence_obj, increases the indentation
        by 1, and prints both of the arguments."""
        self.general_print(sentence_obj, eval_world, indent_num)


class BiconditionalOperator(syntactic.DefinedOperator):

    name = "\\leftrightarrow"
    arity = 2

    def derived_definition(self, leftarg, rightarg):
        right_to_left = [ConditionalOperator, leftarg, rightarg]
        left_to_right = [ConditionalOperator, rightarg, leftarg]
        return [AndOperator, right_to_left, left_to_right]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Prints the proposition for sentence_obj, increases the indentation
        by 1, and prints both of the arguments."""
        self.general_print(sentence_obj, eval_world, indent_num)



##############################################################################
####################### DEFINED CONSTITUTIVE OPERATORS #######################
##############################################################################

class DefGroundOperator(syntactic.DefinedOperator):

    name = "\\ground"
    arity = 2

    def derived_definition(self, leftarg, rightarg):
        return [IdentityOperator, [OrOperator, leftarg, rightarg], rightarg]

    def print_method(self, sentence_obj, eval_world, indent_num):
        """Prints the proposition for sentence_obj, increases the indentation
        by 1, and prints both of the arguments."""
        self.general_print(sentence_obj, eval_world, indent_num)


class DefEssenceOperator(syntactic.DefinedOperator):

    name = "\\essence"
    arity = 2

    def derived_definition(self, leftarg, rightarg):
        return [IdentityOperator, [AndOperator, leftarg, rightarg], rightarg]

    def print_method(self, sentence_obj, eval_world, indent_num):
        """Prints the proposition for sentence_obj, increases the indentation
        by 1, and prints both of the arguments."""
        self.general_print(sentence_obj, eval_world, indent_num)



##############################################################################
###################### DEFINED COUNTERFACTUAL OPERATORS ######################
##############################################################################

class MightCounterfactualOperator(syntactic.DefinedOperator):

    name = "\\circleright"
    arity = 2

    def derived_definition(self, leftarg, rightarg):
        return [
            NegationOperator, [
                CounterfactualOperator,
                leftarg,
                [NegationOperator, rightarg]
            ]
        ]

    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        is_alt = self.semantics.calculate_alternative_worlds
        model_structure = sentence_obj.proposition.model_structure
        left_argument_obj = sentence_obj.original_arguments[0]
        left_argument_verifiers = left_argument_obj.proposition.verifiers
        alt_worlds = is_alt(left_argument_verifiers, eval_world, model_structure)
        self.print_over_worlds(sentence_obj, eval_world, alt_worlds, indent_num)



class MightImpositionOperator(syntactic.DefinedOperator):

    name = "\\could"
    arity = 2

    def derived_definition(self, leftarg, rightarg):
        return [
            NegationOperator, [
                ImpositionOperator,
                leftarg,
                [NegationOperator, rightarg]
            ]
        ]

    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        is_outcome = self.semantics.calculate_outcome_worlds
        model_structure = sentence_obj.proposition.model_structure
        left_argument_obj = sentence_obj.original_arguments[0]
        left_argument_verifiers = left_argument_obj.proposition.verifiers
        alt_worlds = is_outcome(left_argument_verifiers, eval_world, model_structure)
        self.print_over_worlds(sentence_obj, eval_world, alt_worlds, indent_num)



##############################################################################
####################### DEFINED INTENSIONAL OPERATORS ########################
##############################################################################

class DefNecessityOperator(syntactic.DefinedOperator):

    name = "\\necessary"
    arity = 1

    def derived_definition(self, rightarg):
        # NOTE: TopOperator is not a list like the others, so [TopOperator]
        return [CounterfactualOperator, [TopOperator], rightarg]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        all_worlds = sentence_obj.proposition.model_structure.world_bits
        self.print_over_worlds(sentence_obj, eval_world, all_worlds, indent_num)


class DefPossibilityOperator(syntactic.DefinedOperator):

    name = "\\possible"
    arity = 1

    def derived_definition(self, arg):
        return [NegationOperator, [DefNecessityOperator, [NegationOperator, arg]]]
        # return [NegationOperator, [NecessityOperator, [NegationOperator, arg]]]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        all_worlds = sentence_obj.proposition.model_structure.world_bits
        self.print_over_worlds(sentence_obj, eval_world, all_worlds, indent_num)


class DefPossibilityOperator2(syntactic.DefinedOperator):

    name = "\\possible2" # note name change
    arity = 1

    def derived_definition(self, arg):
        return [MightCounterfactualOperator, [TopOperator], arg]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        all_worlds = sentence_obj.proposition.model_structure.world_bits
        self.print_over_worlds(sentence_obj, eval_world, all_worlds, indent_num)


##############################################################################
####################### CIRCULAR DEFINITIONS TESTING #########################
##############################################################################

class CircNecessityOperator(syntactic.DefinedOperator):
    """Defined to check for definitional loops."""

    name = "\\circNec"
    arity = 1

    def derived_definition(self, arg):
        return [NegationOperator, [CircPossibilityOperator, [NegationOperator, arg]]]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        all_worlds = sentence_obj.proposition.model_structure.world_bits
        self.print_over_worlds(sentence_obj, eval_world, all_worlds, indent_num)


class CircPossibilityOperator(syntactic.DefinedOperator):
    """Defined to check for definitional loops."""

    name = "\\circPoss"
    arity = 1

    def derived_definition(self, arg):
        return [NegationOperator, [CircNecessityOperator, [NegationOperator, arg]]]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        all_worlds = sentence_obj.proposition.model_structure.world_bits
        self.print_over_worlds(sentence_obj, eval_world, all_worlds, indent_num)


class CircNecessityOperator1(syntactic.DefinedOperator):
    """Defined to check for definitional loops."""

    name = "\\circNec1"
    arity = 1

    def derived_definition(self, arg):
        return [NegationOperator, [CircPossibilityOperator1, [NegationOperator, arg]]]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        all_worlds = sentence_obj.proposition.model_structure.world_bits
        self.print_over_worlds(sentence_obj, eval_world, all_worlds, indent_num)


class CircPossibilityOperator1(syntactic.DefinedOperator):
    """Defined to check for definitional loops."""

    name = "\\circPoss1"
    arity = 1

    def derived_definition(self, arg):
        return [NegationOperator, [CircNecessityOperator2, [NegationOperator, arg]]]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        all_worlds = sentence_obj.proposition.model_structure.world_bits
        self.print_over_worlds(sentence_obj, eval_world, all_worlds, indent_num)


class CircNecessityOperator2(syntactic.DefinedOperator):
    """Defined to check for definitional loops."""

    name = "\\circNec2"
    arity = 1

    def derived_definition(self, arg):
        return [NegationOperator, [CircPossibilityOperator2, [NegationOperator, arg]]]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        all_worlds = sentence_obj.proposition.model_structure.world_bits
        self.print_over_worlds(sentence_obj, eval_world, all_worlds, indent_num)


class CircPossibilityOperator2(syntactic.DefinedOperator):
    """Defined to check for definitional loops."""

    name = "\\circPoss2"
    arity = 1

    def derived_definition(self, arg):
        return [NegationOperator, [CircNecessityOperator1, [NegationOperator, arg]]]
    
    def print_method(self, sentence_obj, eval_world, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        all_worlds = sentence_obj.proposition.model_structure.world_bits
        self.print_over_worlds(sentence_obj, eval_world, all_worlds, indent_num)



"""run 'pytest' from the '.../Code/' directory"""
import pytest
from test.utils import (
    check_model_status,
    default_max_time,
)

from src.model_checker.syntactic import (
    OperatorCollection,
)

from src.model_checker.semantic import (
    Proposition,
    Semantics,
)

from src.model_checker.defined import (
    NegationOperator,
    AndOperator,
    OrOperator,
    ConditionalOperator,
    BiconditionalOperator,
    TopOperator,
    BotOperator,
)

semantics = Semantics
proposition = Proposition
operators = OperatorCollection(
    NegationOperator,
    AndOperator,
    OrOperator,
    ConditionalOperator,
    BiconditionalOperator,
    TopOperator,
    BotOperator,
)

max_time = default_max_time

#####################################
##### EXTENSIONAL COUNTERMODELS #####
#####################################

@pytest.mark.timeout(max_time)
def test_EXT_CM1():
    premises = ['A']
    conclusions = ['\\neg A']
    settings = {
        'N' : 3,
        'desired_status' : True,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )






#############################
##### EXTENSIONAL LOGIC #####
#############################

@pytest.mark.timeout(max_time)
def test_EXT1():
    premises = ['A','(A \\rightarrow B)']
    conclusions = ['B']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_EXT2():
    premises = []
    conclusions = ['(A \\rightarrow (B \\rightarrow A))']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_EXT3():
    premises = []
    conclusions = ['((A \\rightarrow (B \\rightarrow C)) \\rightarrow ((A \\rightarrow B) \\rightarrow (A \\rightarrow C)))']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_EXT4():
    premises = []
    conclusions = ['((\\neg A \\rightarrow \\neg B) \\rightarrow (B \\rightarrow A))']
    settings = {
        'N' : 3,
        'desired_status' : False,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )



"""run 'pytest' from the '.../Code/src/model_checker/' directory"""
import pytest

from test.utils import (
    check_model_status,
    default_max_time,
)

from src.model_checker.syntactic import (
    OperatorCollection,
)

from src.model_checker.semantic import (
    Proposition,
    ImpositionSemantics,
)

from src.model_checker.defined import (
    NegationOperator,
    AndOperator,
    OrOperator,
    ImpositionOperator,
    PossibilityOperator,
    NecessityOperator,
    MightImpositionOperator,
)

semantics = ImpositionSemantics
proposition = Proposition
operators = OperatorCollection(
    NegationOperator,
    AndOperator,
    OrOperator,
    ImpositionOperator,
    MightImpositionOperator,
    NecessityOperator,
    PossibilityOperator,
)

max_time = default_max_time

########################################
##### IMPOSITION COUNTERMODELS #####
########################################

@pytest.mark.timeout(max_time)
def test_IMP_CM1():
# IMPOSITION ANTECEDENT STRENGTHENING
    premises = ['(A \\imposition C)']
    conclusions = ['((A \\wedge B) \\imposition C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM2():
# MIGHT IMPOSITION ANTECEDENT STRENGTHENING
    premises = ['(A \\could C)']
    conclusions = ['((A \\wedge B) \\could C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM3():
# IMPOSITION ANTECEDENT STRENGTHENING WITH POSSIBILITY
    premises = ['(A \\imposition C)', '\\Diamond (A \\wedge B)']
    conclusions = ['((A \\wedge B) \\imposition C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM4():
# IMPOSITION ANTECEDENT STRENGTHENING WITH NEGATION
    premises = ['\\neg A', '(A \\imposition C)']
    conclusions = ['((A \\wedge B) \\imposition C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM5():
# IMPOSITION DOUBLE ANTECEDENT STRENGTHENING
    premises = ['(A \\imposition C)','(B \\imposition C)']
    conclusions = ['((A \\wedge B) \\imposition C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM6():
# WEAKENED MONOTONICITY
    premises = ['(A \\imposition B)','(A \\imposition C)']
    conclusions = ['((A \\wedge B) \\imposition C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )


@pytest.mark.timeout(max_time)
def test_IMP_CM7():
# IMPOSITION CONTRAPOSITION
    premises = ['(A \\imposition B)']
    conclusions = ['(\\neg B \\imposition \\neg A)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM8():
# IMPOSITION CONTRAPOSITION WITH NEGATION
    premises = ['\\neg B','(A \\imposition B)']
    conclusions = ['(\\neg B \\imposition \\neg A)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM9():
# IMPOSITION CONTRAPOSITION WITH TWO NEGATIONS
    premises = ['\\neg A','\\neg B','(A \\imposition B)']
    conclusions = ['(\\neg B \\imposition \\neg A)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM10():
# TRANSITIVITY
    premises = ['(A \\imposition B)','(B \\imposition C)']
    conclusions = ['(A \\imposition C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM11():
# IMPOSITION TRANSITIVITY WITH NEGATION
    premises = ['\\neg A','(A \\imposition B)','(B \\imposition C)']
    conclusions = ['(A \\imposition C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM12():
# IMPOSITION TRANSITIVITY WITH TWO NEGATIONS
    premises = ['\\neg A','\\neg B','(A \\imposition B)','(B \\imposition C)']
    conclusions = ['(A \\imposition C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM13():
# SOBEL SEQUENCE (N = 3)
    premises = [
        '(A \\imposition X)',
        '\\neg ((A \\wedge B) \\imposition X)',
        '(((A \\wedge B) \\wedge C) \\imposition X)',
    ]
    conclusions = []
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM14():
# SOBEL SEQUENCE WITH POSSIBILITY (N = 4)
    premises = [
        '\\Diamond A',
        '(A \\imposition X)',
        '\\Diamond (A \\wedge B)',
        '\\neg ((A \\wedge B) \\imposition X)',
        '\\Diamond ((A \\wedge B) \\wedge C)',
        '(((A \\wedge B) \\wedge C) \\imposition X)',
        '\\Diamond (((A \\wedge B) \\wedge C) \\wedge D)',
    ]
    conclusions = []
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM15():
# IMPOSITION EXCLUDED MIDDLE
    premises = ['\\neg A']
    conclusions = ['(A \\imposition B)','(A \\imposition \\neg B)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM16():
# SIMPLIFICATION OF DISJUNCTIVE CONSEQUENT
    premises = ['\\neg A','(A \\imposition (B \\vee C))']
    conclusions = ['(A \\imposition B)','(A \\imposition C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM17():
# INTRODUCTION OF DISJUNCTIVE ANTECEDENT
    premises = ['(A \\imposition C)','(B \\imposition C)']
    conclusions = ['((A \\vee B) \\imposition C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM18():
# MUST FACTIVITY
    premises = ['A', 'B']
    conclusions = ['(A \\imposition B)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM19():
# IMPOSITION EXPORTATION
    premises = ['((A \\wedge B) \\imposition C)']
    conclusions = ['(A \\imposition (B \\imposition C))']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM20():
# IMPOSITION EXPORTATION WITH POSSIBILITY
    premises = ['((A \\wedge B) \\imposition C)','\\Diamond (A \\wedge B)']
    conclusions = ['(A \\imposition (B \\imposition C))']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_CM21():
# IMPOSITION EXCLUDED MIDDLE VARIANT
    premises = ['\\neg A','\\neg (A \\imposition B)']
    conclusions = ['(A \\imposition \\neg B)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

# # NOTE: DOES NOT FIND COUNTERMODEL
# @pytest.mark.timeout(max_time)
# def test_CL_8():
#     premises = ['(A \\imposition (B \\imposition C))']
#     conclusions = ['((A \\wedge B) \\imposition C)']
#     settings = {
#         'N' : 4,
#         'contingent' : True,
#         'non_null' : True,
#         'non_empty' : True,
#         'disjoint' : False,
#         'desired_status' : True,
#         'print_impossible' : True,
#         'max_time' : max_time,
#     }
#     check_model_status(
#         premises,
#         conclusions,
#         semantics,
#         proposition,
#         operators,
#         settings,
#     )





################################
##### IMPOSITION LOGIC #####
################################

@pytest.mark.timeout(max_time)
def test_IMP_T1():
    """IMPOSITION IDENTITY"""
    premises = []
    conclusions = ['(A \\imposition A)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_T2():
    """IMPOSITION MODUS PONENS"""
    premises = ['A','(A \\imposition B)']
    conclusions = ['B']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_T3():
    """WEAKENED TRANSITIVITY"""
    premises = ['(A \\imposition B)','((A \\wedge B) \\imposition C)']
    conclusions = ['(A \\imposition C)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_T4():
    """ANTECEDENT DISJUNCTION TO CONJUNCTION"""
    premises = ['((A \\vee B) \\imposition C)']
    conclusions = ['((A \\wedge B) \\imposition C)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_T5():
    """SIMPLIFICATION OF DISJUNCTIVE ANTECEDENT"""
    premises = ['((A \\vee B) \\imposition C)']
    conclusions = ['(A \\imposition C)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_T6():
    """DOUBLE SIMPLIFICATION OF DISJUNCTIVE ANTECEDENT"""
    premises = ['((A \\vee B) \\imposition C)']
    conclusions = ['((A \\imposition C) \\wedge (B \\imposition C))']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_T7():
    premises = ['(A \\imposition C)', '(B \\imposition C)', '((A \\wedge B) \\imposition C)']
    conclusions = ['((A \\vee B) \\imposition C)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_T8():
    premises = ['(A \\imposition (B \\wedge C))']
    conclusions = ['(A \\imposition B)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_T9():
    premises = ['(A \\imposition B)','(A \\imposition C)']
    conclusions = ['(A \\imposition (B \\wedge C))']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_IMP_T10():
    """FACTIVITY MIGHT"""
    premises = ['A','B']
    conclusions = ['(A \\could B)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )



'''
file specifies premises, conclusions, and settings.
running the file finds a model and prints the result.
to test from the Code/ directory, run: python -m src.model_checker 
'''

import sys
import os
import shutil
import subprocess
import argparse
import importlib.util
from model_checker.__init__ import __version__

class BuildModule:
    """load module and store values as a class"""
    def __init__(self, module_name, module_path):
        self.module_name = module_name
        self.module_path = module_path
        self.module = self.load_module()
        self.default_values = {
            "N": 3,
            "premises": [],
            "conclusions": [],
            "max_time": 1,
            "contingent": False,
            "non_empty": False,
            "non_null": False,
            "disjoint": False,
            # "optimize": False,
            "print_constraints": False,
            "print_impossible": False,
            "save_output": False,
        }
        self.settings = getattr(
            self.module,
            "settings",
            None
        )
        self.model_structure = getattr(
            self.module,
            "model_structure",
            None
        )
        self.print_command = getattr(
            self.module,
            "print_command",
            None
        )
        self.N = getattr(
            self.settings,
            "N",
            None
        )
        self.premises = getattr(
            self.module,
            "premises",
            self.default_values["premises"]
        )
        self.conclusions = getattr(
            self.module,
            "conclusions",
            self.default_values["conclusions"]
        )
        self.max_time = getattr(
            self.settings,
            "max_time",
            self.default_values["max_time"]
        )
        self.contingent = getattr(
            self.settings,
            "contingent",
            self.default_values["contingent"]
        )
        self.non_empty = getattr(
            self.settings,
            "non_empty",
            self.default_values["non_empty"]
        )
        self.non_null = getattr(
            self.settings,
            "non_null",
            self.default_values["non_null"]
        )
        self.disjoint = getattr(
            self.settings,
            "disjoint",
            self.default_values["disjoint"]
        )
        # self.optimize = getattr(
        #     self.settings,
        #     "optimize",
        #     self.default_values["optimize"]
        # )
        self.print_constraints = getattr(
            self.settings,
            "print_constraints",
            self.default_values["print_constraints"]
        )
        self.print_impossible = getattr(
            self.settings,
            "print_impossible",
            self.default_values["print_impossible"]
        )
        self.save_output = getattr(
            self.module,
            "save_output",
            self.default_values["save_output"]
        )
        # TODO: fix
        # self.validate_attributes()

    def load_module(self):
        """prepares a test file, raising a error if unsuccessful."""
        try:
            spec = importlib.util.spec_from_file_location(self.module_name, self.module_path)
            if spec is None or spec.loader is None:
                raise ImportError("Module spec could not be loaded.")
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module
        except Exception as e:
            raise ImportError(f"Failed to load the module '{self.module_name}': {e}") from e

    # TODO: fix to check just the settings
    # def validate_attributes(self):
    #     # print(f"TEST: {self.settings}")
    #     for attr, default_value in self.default_values.items():
    #         # print(f"ATTR {attr} {type(attr)} DEFAULT {default_value} {type(default_value)}")
    #         if attr not in self.settings:
    #             print(f"The value of '{attr}' is absent and has been set to {default_value}.")

def parse_file_and_flags():
    """returns the name and path for the current script"""
    # create an ArgumentParser object
    parser = argparse.ArgumentParser(
        prog='model-checker',
        usage='%(prog)s [options] input',
        description="""
        Running '%(prog)s' without options or an input will prompt the user
        to generate a new test file. To run a test on an existing file, include
        the path to that file as the input.""",
        epilog="""
        More information can be found at:
        https://github.com/benbrastmckie/ModelChecker/""",
    )
    parser.add_argument(
        "file_path",
        nargs='?',
        default=None,
        type=str,
        help="Specifies the path to a Python.",
    )
    parser.add_argument(
        '--contingent',
        '-c',
        action='store_true',
        help='Overrides to make all propositions contingent.'
    )
    parser.add_argument(
        '--non_empty',
        '-e',
        action='store_true',
        help='Overrides to make all propositions non_empty.'
    )
    parser.add_argument(
        '--non_null',
        '-n',
        action='store_true',
        help='Overrides to make all propositions non_null.'
    )
    parser.add_argument(
        '--disjoint',
        '-d',
        action='store_true',
        help='Overrides to make all propositions have disjoint subject-matters.'
    )
    parser.add_argument(
        '--print',
        '-p',
        action='store_true',
        help='Overrides to print the Z3 constraints or else the unsat_core constraints if there is no model.'
    )
    parser.add_argument(
        '--save_output',
        '-s',
        action='store_true',
        help='Overrides to prompt user to save output.'
    )
    parser.add_argument(
        '--impossible',
        '-i',
        action='store_true',
        help='Overrides to print impossible states.'
    )
    parser.add_argument(
        '--version',
        '-v',
        action='version',
        version=f"%(prog)s:  {__version__}",
        help='Prints the version number.'
    )
    # parser.add_argument(
    #     '--optimize',
    #     '-o',
    #     action='store_true',
    #     help='finds the minimum value for N that is satisfiable before reaching max_time.'
    # )
    parser.add_argument(
        '--upgrade',
        '-u',
        action='store_true',
        help='Upgrade the package.'
    )
    # parse the command-line argument to get the module path
    args = parser.parse_args()
    package_name = parser.prog  # Get the package name from the parser
    return args, package_name

# def generate_test(name):
#     """check if a script name was provided"""
#     template_data = {
#         'name': name
#     }
#     script_content = script_template.substitute(template_data)
#     output_file_path = f'{name}.py'
#     with open(output_file_path, 'w', encoding="utf-8") as f:
#         f.write(script_content)
#     print(f"\nThe {name}.py file has been created.")
#     print("You can run this file with the command:\n")
#     print(f"model-checker {name}.py\n")

def generate_project(name):
    """
    Copy the 'template/' directory to the current working directory, 
    rename it to the specified 'name', and rename its modules by prefixing 'name'.
    """
    project_name = 'project_' + name
    # Dynamically resolve the source directory relative to this script
    source_dir = os.path.join(os.path.dirname(__file__), 'template')  
    destination_dir = os.path.join(os.getcwd(), project_name)  

    try:
        # Check if the source directory exists
        if not os.path.exists(source_dir):
            raise FileNotFoundError(f"The source directory '{source_dir}' was not found.")

        # Check if the destination directory already exists
        if os.path.exists(destination_dir):
            print(f"Error: Directory '{destination_dir}' already exists.")
            return

        # Copy the template directory
        shutil.copytree(source_dir, destination_dir)

        # Rename the files in the copied directory
        files_to_rename = {
            "examples.py": "examples.py",
            "operators.py": "operators.py",
            "semantics.py": "semantics.py",
        }
        
        for old_name, new_name in files_to_rename.items():
            old_path = os.path.join(destination_dir, old_name)
            new_path = os.path.join(destination_dir, new_name)
            if os.path.exists(old_path):
                os.rename(old_path, new_path)

        print(f"\nProject generated at: {destination_dir}\n")
        print("The following modules were created:")
        for old_name, new_name in files_to_rename.items():
            print(f"  {new_name}")

        # Run the example script
        example_script = os.path.join(destination_dir, "examples.py")
        print(example_script)
        if os.path.exists(example_script):
            print("\nRunning the example script...")
            subprocess.run(["model-checker", example_script])
        else:
            print(f"\nFail to run: model-checker {example_script}")

        # Output the test command
        print(f"\nRun the following command to test your project:\n\nmodel-checker {destination_dir}/examples.py\n")
    
    except FileNotFoundError as e:
        print(f"Error: {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")

def ask_generate_project():
    """prompt user to create a test file"""
    result = input("Would you like to generate a new project? (y/n): ")
    if result in {'Yes', 'yes', 'Ye', 'ye', 'Y', 'y'}:
        test_name = input("Enter the name of your project using snake_case: ")
        generate_project(test_name)
        return
    print("You can run an example.py file that already exists with the command:\n")
    print("model-checker example.py\n")

def ask_save():
    """print the model and prompt user to store the output"""
    result = input("Would you like to save the output? (y/n):\n")
    if not result in ['Yes', 'yes', 'Ye', 'ye', 'Y', 'y']:
        VOID = ""
        return VOID, False
    cons_input = input("\nWould you like to include the Z3 constraints? (y/n):\n")
    print_cons = bool(cons_input in ['Yes', 'yes', 'Ye', 'ye', 'Y', 'y'])
    file_name = input(
        "\nEnter the file name in snake_case without an extension.\n"
        "Leave the file name blank to append the output to the project file.\n"
        "\nFile name: "
    )
    return file_name, print_cons

def no_model_save_or_append(module, model_structure, file_name, print_cons):
    """option to save or append if no model is found"""
    if len(file_name) == 0:
        with open(f"{module.module_path}", 'a', encoding="utf-8") as f:
            print('\n"""', file=f)
            # TODO: add method
            model_structure.no_model_print(print_cons, f)
            print('"""', file=f)
        return
    destination_dir = os.path.join(os.getcwd(), module.module_name)  
    with open(f"{destination_dir}/{file_name}.py", 'w', encoding="utf-8") as n:
        print(f'# TITLE: {file_name}.py created in {destination_dir}\n"""', file=n)
        # TODO: add method
        model_structure.no_model_save(print_cons, n)
    print()

def save_or_append(module, model_structure, file_name, print_cons):
    """Option to save or append if a model is found."""
    if len(file_name) == 0:
        with open(f"{module.module_path}", 'a', encoding="utf-8") as f:
            print('\n"""', file=f)
            model_structure.print_to(print_cons, f)  # TODO: add function
            print('"""', file=f)
        return

    # Default or fallback path
    project_name = getattr(module, 'module_name', 'project')
    destination_dir = os.path.join(os.getcwd(), project_name)
    
    # Ensure the directory exists
    os.makedirs(destination_dir, exist_ok=True)

    with open(f"{destination_dir}/{file_name}.py", 'w', encoding="utf-8") as n:
        print(f'# TITLE: {file_name}.py created in {destination_dir}\n"""', file=n)
        model_structure.save_to(print_cons, n)  # TODO: add function
    print(f'\n{file_name}.py created in {destination_dir}\n')

def create_module(args):
    """Returns a module from the arguments provided from the specified file.
    Updates the model to reflect the user specified flags."""
    module_path = args.file_path
    module_name = os.path.splitext(os.path.basename(module_path))[0]
    module = BuildModule(module_name, module_path)
    module.contingent = module.contingent or args.contingent
    module.non_empty = module.non_empty or args.non_empty
    module.non_null = module.non_null or args.non_null
    module.disjoint = module.disjoint or args.disjoint
    # module.optimize = module.optimize or args.optimize
    module.print_constraints = module.print_constraints or args.print
    module.print_impossible = module.print_impossible or args.impossible
    module.save_output = module.save_output or args.save_output
    return module

def print_result(module):
    """Prints resulting model or no model if none is found."""
    model_structure = module.model_structure
    if model_structure.z3_model_status:
        module.print_command
        # TODO: turn on once print_to and save_to have been added
        if module.save_output:
            file_name, print_cons = ask_save()
            save_or_append(module, model_structure, file_name, print_cons)
        return
    # # TODO: turn on once print_to and save_to have been added
    # model_structure.no_model_print(module.print_constraints)
    # if module.save_output:
    #     file_name, print_cons = ask_save()
    #     no_model_save_or_append(module, model_structure, file_name, print_cons)


def main():
    if len(sys.argv) < 2:
        ask_generate_project()
        return
    args, package_name = parse_file_and_flags()
    if args.upgrade:
        print("Upgrading package")
        try:
            subprocess.run(['pip', 'install', '--upgrade', package_name], check=True)
        except subprocess.CalledProcessError as e:
            print(f"Failed to upgrade {package_name}: {e}")
        return

    module = create_module(args)

    print_result(module)

if __name__ == "__main__":
    main()



"""From the project directory, run: python -m src.model_checker.template.examples"""

import sys
import os
sys.path.append(os.path.dirname(__file__))  # Add the current directory to sys.path

from semantic import (
    Semantics,
    Proposition,
)

from operators import (
    operators,
)

from model_checker import syntactic

from model_checker import model

########################
### DEFAULT SETTINGS ###
########################

settings = {
    'N' : 3,
    'contingent' : True,
    'non_empty' : True,
    'non_null' : True,
    'disjoint' : False,
    'print_impossible' : True,
    'max_time' : 1,
}



#####################
### COUNTERMODELS ###
#####################

# # CF_CM1: COUNTERFACTUAL ANTECEDENT STRENGTHENING
CF_CM1_premises = ['(A \\boxright C)']
CF_CM1_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM1_example = [
    CF_CM1_premises,
    CF_CM1_conclusions,
]

# # CF_CM2: MIGHT COUNTERFACTUAL ANTECEDENT STRENGTHENING
CF_CM2_premises = ['(A \\circleright C)']
CF_CM2_conclusions = ['((A \\wedge B) \\circleright C)']
CF_CM2_example = [
    CF_CM2_premises,
    CF_CM2_conclusions,
]

# # CF_CM7: COUNTERFACTUAL CONTRAPOSITION
CF_CM7_premises = ['(A \\boxright B)']
CF_CM7_conclusions = ['(\\neg B \\boxright \\neg A)']
CF_CM7_example = [
    CF_CM7_premises,
    CF_CM7_conclusions,
]

# # CF_CM10: TRANSITIVITY
CF_CM10_premises = ['(A \\boxright B)','(B \\boxright C)']
CF_CM10_conclusions = ['(A \\boxright C)']
CF_CM10_example = [
    CF_CM10_premises,
    CF_CM10_conclusions,
]

# # CF_CM13: SOBEL SEQUENCE
CF_CM13_premises = [
    '(A \\boxright X)',
    '\\neg ((A \\wedge B) \\boxright X)',
    '(((A \\wedge B) \\wedge C) \\boxright X)',
    '\\neg ((((A \\wedge B) \\wedge C) \\wedge D) \\boxright X)',
    '(((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\boxright X)',
    '\\neg ((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\boxright X)',
    '(((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\wedge G) \\boxright X)', # 327.2 seconds on the MIT servers; now .01244 seconds
]
CF_CM13_conclusions = []
CF_CM13_example = [
    CF_CM13_premises,
    CF_CM13_conclusions,
]

# # CF_CM19: COUNTERFACTUAL EXPORTATION
CF_CM19_premises = ['((A \\wedge B) \\boxright C)']
CF_CM19_conclusions = ['(A \\boxright (B \\boxright C))']
CF_CM19_example = [
    CF_CM19_premises,
    CF_CM19_conclusions,
]


############################
### LOGICAL CONSEQUENCES ###
############################

# # CF_T2: COUNTERFACTUAL MODUS PONENS
CF_T2_premises = ['A','(A \\boxright B)']
CF_T2_conclusions = ['B']
CF_T2_example = [
    CF_T2_premises,
    CF_T2_conclusions
]

# # CF_T3: WEAKENED TRANSITIVITY
CF_T3_premises = ['(A \\boxright B)','((A \\wedge B) \\boxright C)']
CF_T3_conclusions = ['(A \\boxright C)']
CF_T3_example = [
    CF_T3_premises,
    CF_T3_conclusions
]


# # CF_T5: SIMPLIFICATION OF DISJUNCTIVE ANTECEDENT
CF_T5_premises = ['((A \\vee B) \\boxright C)']
CF_T5_conclusions = ['(A \\boxright C)']
CF_T5_example = [
    CF_T5_premises,
    CF_T5_conclusions
]

# # CF_T_T11: DEFINITION OF NEC
CF_T11_premises = ['(\\neg A \\boxright \\bot)']
CF_T11_conclusions = ['(\\top \\boxright A)']
CF_T11_example = [
    CF_T11_premises,
    CF_T11_conclusions,
]

#########################################
### GENERATE Z3 CONSTRAINTS AND PRINT ###
#########################################

premises, conclusions = CF_CM1_example

syntax = syntactic.Syntax(premises, conclusions, operators)

semantics = Semantics(settings['N'])

model_constraints = model.ModelConstraints(
    settings,
    syntax,
    semantics,
    Proposition,
)

model_structure = model.ModelStructure(model_constraints)

print_command = model_structure.print_all()
# model_structure.print_all()





"""run 'pytest' from the '.../Code/' directory"""
import pytest
from test.utils import (
    check_model_status,
    default_max_time,
)

from src.model_checker.syntactic import (
    OperatorCollection,
)

from src.model_checker.semantic import (
    Proposition,
    Semantics,
)

from src.model_checker.defined import (
    NegationOperator,
    AndOperator,
    OrOperator,
    CounterfactualOperator,
    PossibilityOperator,
    NecessityOperator,
    MightCounterfactualOperator,
)

semantics = Semantics
proposition = Proposition
operators = OperatorCollection(
    NegationOperator,
    AndOperator,
    OrOperator,
    CounterfactualOperator,
    MightCounterfactualOperator,
    NecessityOperator,
    PossibilityOperator,
)

max_time = default_max_time

########################################
##### COUNTERFACTUAL COUNTERMODELS #####
########################################

@pytest.mark.timeout(max_time)
def test_CF_CM1():
# COUNTERFACTUAL ANTECEDENT STRENGTHENING
    premises = ['(A \\boxright C)']
    conclusions = ['((A \\wedge B) \\boxright C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM2():
# MIGHT COUNTERFACTUAL ANTECEDENT STRENGTHENING
    premises = ['(A \\circleright C)']
    conclusions = ['((A \\wedge B) \\circleright C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM3():
# COUNTERFACTUAL ANTECEDENT STRENGTHENING WITH POSSIBILITY
    premises = ['(A \\boxright C)', '\\Diamond (A \\wedge B)']
    conclusions = ['((A \\wedge B) \\boxright C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM4():
# COUNTERFACTUAL ANTECEDENT STRENGTHENING WITH NEGATION
    premises = ['\\neg A', '(A \\boxright C)']
    conclusions = ['((A \\wedge B) \\boxright C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM5():
# COUNTERFACTUAL DOUBLE ANTECEDENT STRENGTHENING
    premises = ['(A \\boxright C)','(B \\boxright C)']
    conclusions = ['((A \\wedge B) \\boxright C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM6():
# WEAKENED MONOTONICITY
    premises = ['(A \\boxright B)','(A \\boxright C)']
    conclusions = ['((A \\wedge B) \\boxright C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )


@pytest.mark.timeout(max_time)
def test_CF_CM7():
# COUNTERFACTUAL CONTRAPOSITION
    premises = ['(A \\boxright B)']
    conclusions = ['(\\neg B \\boxright \\neg A)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM8():
# COUNTERFACTUAL CONTRAPOSITION WITH NEGATION
    premises = ['\\neg B','(A \\boxright B)']
    conclusions = ['(\\neg B \\boxright \\neg A)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM9():
# COUNTERFACTUAL CONTRAPOSITION WITH TWO NEGATIONS
    premises = ['\\neg A','\\neg B','(A \\boxright B)']
    conclusions = ['(\\neg B \\boxright \\neg A)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM10():
# TRANSITIVITY
    premises = ['(A \\boxright B)','(B \\boxright C)']
    conclusions = ['(A \\boxright C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM11():
# COUNTERFACTUAL TRANSITIVITY WITH NEGATION
    premises = ['\\neg A','(A \\boxright B)','(B \\boxright C)']
    conclusions = ['(A \\boxright C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM12():
# COUNTERFACTUAL TRANSITIVITY WITH TWO NEGATIONS
    premises = ['\\neg A','\\neg B','(A \\boxright B)','(B \\boxright C)']
    conclusions = ['(A \\boxright C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM13():
# SOBEL SEQUENCE (N = 3)
    premises = [
        '(A \\boxright X)',
        '\\neg ((A \\wedge B) \\boxright X)',
        '(((A \\wedge B) \\wedge C) \\boxright X)',
    ]
    conclusions = []
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM14():
# SOBEL SEQUENCE WITH POSSIBILITY (N = 4)
    premises = [
        '\\Diamond A',
        '(A \\boxright X)',
        '\\Diamond (A \\wedge B)',
        '\\neg ((A \\wedge B) \\boxright X)',
        '\\Diamond ((A \\wedge B) \\wedge C)',
        '(((A \\wedge B) \\wedge C) \\boxright X)',
        '\\Diamond (((A \\wedge B) \\wedge C) \\wedge D)',
    ]
    conclusions = []
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM15():
# COUNTERFACTUAL EXCLUDED MIDDLE
    premises = ['\\neg A']
    conclusions = ['(A \\boxright B)','(A \\boxright \\neg B)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM16():
# SIMPLIFICATION OF DISJUNCTIVE CONSEQUENT
    premises = ['\\neg A','(A \\boxright (B \\vee C))']
    conclusions = ['(A \\boxright B)','(A \\boxright C)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM17():
# INTRODUCTION OF DISJUNCTIVE ANTECEDENT
    premises = ['(A \\boxright C)','(B \\boxright C)']
    conclusions = ['((A \\vee B) \\boxright C)']
    settings = {
        'N' : 4,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM18():
# MUST FACTIVITY
    premises = ['A', 'B']
    conclusions = ['(A \\boxright B)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM19():
# COUNTERFACTUAL EXPORTATION
    premises = ['((A \\wedge B) \\boxright C)']
    conclusions = ['(A \\boxright (B \\boxright C))']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM20():
# COUNTERFACTUAL EXPORTATION WITH POSSIBILITY
    premises = ['((A \\wedge B) \\boxright C)','\\Diamond (A \\wedge B)']
    conclusions = ['(A \\boxright (B \\boxright C))']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_CM21():
# COUNTERFACTUAL EXCLUDED MIDDLE VARIANT
    premises = ['\\neg A','\\neg (A \\boxright B)']
    conclusions = ['(A \\boxright \\neg B)']
    settings = {
        'N' : 3,
        'contingent' : True,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : True,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

# # NOTE: DOES NOT FIND COUNTERMODEL
# @pytest.mark.timeout(max_time)
# def test_CF_CM22():
#     premises = ['(A \\boxright (B \\boxright C))']
#     conclusions = ['((A \\wedge B) \\boxright C)']
#     N = 4
#     contingent = True
#     non_null = True
#     non_null = True
#     disjoint = False
#     desired_status = True
#     check_model_status(
#         premises,
#         conclusions,
#         semantics,
#         proposition,
#         N,
#         contingent,
#         non_null,
#         disjoint,
#         max_time,
#         desired_status,
#     )





################################
##### COUNTERFACTUAL LOGIC #####
################################

@pytest.mark.timeout(max_time)
def test_CF_T1():
    """COUNTERFACTUAL IDENTITY"""
    premises = []
    conclusions = ['(A \\boxright A)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_T2():
    """COUNTERFACTUAL MODUS PONENS"""
    premises = ['A','(A \\boxright B)']
    conclusions = ['B']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_T3():
    """WEAKENED TRANSITIVITY"""
    premises = ['(A \\boxright B)','((A \\wedge B) \\boxright C)']
    conclusions = ['(A \\boxright C)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_T4():
    """ANTECEDENT DISJUNCTION TO CONJUNCTION"""
    premises = ['((A \\vee B) \\boxright C)']
    conclusions = ['((A \\wedge B) \\boxright C)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_T5():
    """SIMPLIFICATION OF DISJUNCTIVE ANTECEDENT"""
    premises = ['((A \\vee B) \\boxright C)']
    conclusions = ['(A \\boxright C)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_T6():
    """DOUBLE SIMPLIFICATION OF DISJUNCTIVE ANTECEDENT"""
    premises = ['((A \\vee B) \\boxright C)']
    conclusions = ['((A \\boxright C) \\wedge (B \\boxright C))']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_T7():
    premises = ['(A \\boxright C)', '(B \\boxright C)', '((A \\wedge B) \\boxright C)']
    conclusions = ['((A \\vee B) \\boxright C)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_T8():
    premises = ['(A \\boxright (B \\wedge C))']
    conclusions = ['(A \\boxright B)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_T9():
    premises = ['(A \\boxright B)','(A \\boxright C)']
    conclusions = ['(A \\boxright (B \\wedge C))']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )

@pytest.mark.timeout(max_time)
def test_CF_T10():
    """MIGHT FACTIVITY"""
    premises = ['A','B']
    conclusions = ['(A \\circleright B)']
    settings = {
        'N' : 3,
        'contingent' : False,
        'non_null' : True,
        'non_empty' : True,
        'disjoint' : False,
        'desired_status' : False,
        'print_impossible' : True,
        'max_time' : max_time,
    }
    check_model_status(
        premises,
        conclusions,
        semantics,
        proposition,
        operators,
        settings,
    )



"""
This module defines the following classes, culminating in the Syntax class
which is passed to ModelConstraints:
    - **Sentence:** this class is responsible for carrying all values relevant to
    each infix_sentence input by the user in the premises or conclusions.
    - **Operator:** this class sets a number of general defaults that are used by
    the operators the user defines as well as the DefinedOperator class.
    - **DefinedOperator:** ... TODO
    - **OperatorCollection:** this class is responsible for storing all user
    defined operators, passing this collection of operators to Syntax.
    - **Syntax:** this class is responsible for generating a dictionary of sentence
    objects for all premises, conclusions, and their subsentences. By drawing
    on the input operator_collection, all sentences in the dictionary are
    initialized to generate and store original_types for each.
"""

from .utils import (
    bitvec_to_substates,
    not_implemented_string,
    flatten,
    parse_expression,
    pretty_set_print,
)

import time

import inspect

from z3 import Const, DeclareSort, ExprRef

AtomSort = DeclareSort("AtomSort")

class Sentence:
    """Given an infix_sentence as input, an instance of this class store the
    original infix_sentence which is used to name the class instance, as well
    as converting and storing that infix_sentence as a prefix_sentence. The
    class instance is later updated in: (1) Syntax to store a original_type which
    depends on an operator_collection; (2) ModelConstraints to store a
    derived_object and operator which depend on the semantics; and (3)
    ModelStructure to store a proposition for the sentence."""

    def __init__(self, infix_sentence):
        
        # store input, prefix string, complexity, and sentences for arguments
        self.name = infix_sentence
        self.prefix_sentence, self.complexity = self.prefix(infix_sentence)

        # recursive clause: initially stores infix_arguments and infix_operator
        # updated by initialize_sentences in Syntax with operator_collection
        # updated by instantiate in ModelConstraints with semantics
        if self.complexity > 0: 
            self.original_arguments = [ # store infix_arguments
                self.infix(arg)
                for arg in self.prefix_sentence[1:]
            ]
            # updated by update_types:
            self.original_operator = self.prefix_sentence[0]
        else:
            self.original_arguments = None
            self.original_operator = None
        self.arguments = None # updated by initialize_types
        self.operator = None # updated by update_types

        # update_types via initialize_sentences in Syntax: operator_collection
        self.sentence_letter = None

        # update_proposition via interpret in ModelStructure: z3_model
        self.proposition = None

    def __str__(self):
        return self.name
    
    def __repr__(self):
        return self.name

    def prefix(self, infix_sentence):
        """For converting from infix to prefix notation without knowing which
        which operators the language includes."""
        tokens = infix_sentence.replace("(", " ( ").replace(")", " ) ").split()
        derived_object, complexity = parse_expression(tokens)
        return derived_object, complexity

    def infix(self, prefix):
        """Takes a sentence in prefix notation (in any of the three kinds)
        and translates it to infix notation (a string)."""

        if hasattr(prefix, 'name'):
            return prefix.name
        if isinstance(prefix, ExprRef):
            return str(prefix)
        if isinstance(prefix, str):
            return prefix
        if isinstance(prefix, (list, tuple)):
            if len(prefix) == 1:
                return self.infix(prefix[0])
            operator = prefix[0]
            op_str = self.infix(operator)
            if len(prefix) == 2:
                return f"{op_str} {self.infix(prefix[1])}"
            left_expr, right_expr = prefix[1], prefix[2]
            return f"({self.infix(left_expr)} {op_str} {self.infix(right_expr)})"
        raise TypeError(f"The prefix {prefix} has a type error in infix().")

    def update_types(self, operator_collection): # used in Syntax
        """Draws on the operator_collection to apply the operators that occur
        in the prefix_sentence in order to generate a original_type which has
        operator classes in place of operator strings and AtomSorts in place
        of sentence letters. If the operator is not primitive, then ."""

        def primitive_operator(original_type):
            """Checks if the main operator is primitive."""
            return getattr(original_type[0], "primitive", True)

        def derive_type(original_type):
            """This function translates a original_type possibly with defined
            operators to a derived_type without defined operators."""
            if primitive_operator(original_type):
                return original_type
            operator, args = original_type[0], original_type[1:]
            if not hasattr(operator, "primitive"):
                raise TypeError(f"Operator {operator} is not a subclass of Operator.")
            if isinstance(operator, type):
                derivation =  operator('a').derived_definition(*args)
            else:
                raise TypeError(
                    f"The operator {operator} is of type {type(operator)} " +
                    f"which is not an operator class."
                )
            return derive_type(derivation)

        def store_types(derived_type):
            if self.name.isalnum(): # sentence letter
                return None, None, derived_type[0]
            if self.name in {'\\top', '\\bot'}: # extremal operator
                return derived_type[0], None, None
            if len(derived_type) > 1: # complex sentence
                operator_type, argument_types = derived_type[0], derived_type[1:]
                infix_arguments = [self.infix(arg_type) for arg_type in argument_types]
                return operator_type, infix_arguments, None
            raise ValueError(f"the derived_type for {self} is invalid in store_types().")

        original_type = operator_collection.apply_operator(self.prefix_sentence)
        if self.original_operator:
            self.original_operator = original_type[0]
        derived_type = derive_type(original_type)
        self.operator, self.arguments, self.sentence_letter = store_types(derived_type)
    
    def update_objects(self, model_constraints): # used in ModelConstraints init
        """Given an instance of ModelConstraints, this method updates the values
        of self.derived_object and self.operator with the semantics that
        model_constraints includes."""

        def activate_operator(some_type):
            if some_type is None: # operator is None if sentence_letter
                return None
            op_dict = model_constraints.operators
            # # NOTE: required for update operator_collection strategy
            # op_dict = model_constraints.operator_collection.operator_dictionary
            return op_dict[some_type.name]

        self.original_operator = activate_operator(self.original_operator)
        self.operator = activate_operator(self.operator)

    def update_proposition(self, model_structure): # used in ModelStructure init
        """Builds a proposition object for the sentence given the model_structure."""
        self.proposition = model_structure.proposition_class(self, model_structure)


class Operator:
    """Defaults inherited by every operator."""

    name = None
    arity = None
    primitive = True

    def __init__(self, semantics):
        op_class = self.__class__.__name__
        if self.__class__ == Operator:
            raise NotImplementedError(not_implemented_string(op_class))
        if self.name is None or self.arity is None:
            raise NameError(
                f"Your operator class {op_class} is missing a name or an arity. "
                + f"Please add them as class properties of {op_class}."
            )
        self.semantics = semantics

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return str(self.name)

    def __eq__(self, other):
        if isinstance(other, Operator):
            return self.name == other.name and self.arity == other.arity
        return False

    def __hash__(self):
        return hash((self.name, self.arity))

    def general_print(self, sentence_obj, eval_world, indent_num):
        proposition = sentence_obj.proposition
        model_structure = proposition.model_structure

        proposition.print_proposition(eval_world, indent_num)
        indent_num += 1

        for arg in sentence_obj.original_arguments:
            model_structure.recursive_print(arg, eval_world, indent_num)

    def print_over_worlds(self, sentence_obj, eval_world, other_worlds, indent_num):
        """Print counterfactual and the antecedent in the eval_world. Then
        print the consequent in each alternative to the evaluation world.
        """
        # Move to class or config for flexibility
        CYAN, RESET = '\033[36m', '\033[0m'

        arguments = sentence_obj.original_arguments
        proposition = sentence_obj.proposition
        model_structure = proposition.model_structure
        N = proposition.N

        proposition.print_proposition(eval_world, indent_num)
        indent_num += 1

        if len(arguments) == 1:
            argument = arguments[0]
            for world in other_worlds:
                model_structure.recursive_print(argument, world, indent_num)
        if len(arguments) == 2:
            left_argument, right_argument = arguments
            model_structure.recursive_print(left_argument, eval_world, indent_num)
            indent_num += 1
            other_world_strings = {bitvec_to_substates(u, N) for u in other_worlds}
            print(
                f'{"  " * indent_num}{CYAN}|{left_argument}|-alternatives '
                f'to {bitvec_to_substates(eval_world, N)} = '
                f'{pretty_set_print(other_world_strings)}{RESET}'
            )
            indent_num += 1
            for alt_world in other_worlds:
                model_structure.recursive_print(right_argument, alt_world, indent_num)
    
class DefinedOperator(Operator):
    """Represents an operator defined in terms of other operators."""

    primitive = False

    def derived_definition(self, *args):
        """
        Returns the definition of the operator in terms of other operators.
        Must be implemented by subclasses.
        """
        raise NotImplementedError(
            f"Derived operator class {self.__class__.__name__} must implement the derived_definition method."
        )

    def __init__(self, semantics):
        super().__init__(semantics)
        self._validate_arity()

    def _validate_arity(self):
        """
        Validates that the 'arity' attribute matches the number of arguments
        in the 'derived_definition' method (excluding 'self').
        """
        # Retrieve the signature of the derived_definition method
        signature = inspect.signature(self.derived_definition)
        params = list(signature.parameters.values())

        # Exclude 'self' if present
        if params and params[0].name == 'self':
            params = params[1:]

        derived_def_num_args = len(params)

        # Check if 'arity' is defined
        if not hasattr(self, 'arity'):
            raise AttributeError(
                f"{self.__class__.__name__} must define an 'arity' attribute."
            )

        # Validate that 'arity' matches the number of arguments
        if self.arity != derived_def_num_args:
            raise ValueError(
                f"The specified arity of {self.arity} for {self.__class__.__name__} does not match "
                f"the number of arguments ({derived_def_num_args}) for its 'derived_definition' method."
            )


class OperatorCollection:
    """Stores the operators that will be passed to Syntax."""

    def __init__(self, *input):
        self.operator_dictionary = {}
        if input:
            self.add_operator(input)

    def __iter__(self):
        yield from self.operator_dictionary

    def __getitem__(self, value):
        return self.operator_dictionary[value]
    
    def items(self):
        yield from self.operator_dictionary.items()

    def add_operator(self, input):
        """Input is either an operator class (of type 'type') or a list/tuple of operator classes."""
        if isinstance(input, (list, tuple, set)):
            for operator_class in input:
                self.add_operator(operator_class)
        elif isinstance(input, type):
            if input.name in self.operator_dictionary.keys():
                return
            if getattr(input, "name", None) is None:
                raise ValueError(f"Operator class {input.__name__} has no name defined.")
            self.operator_dictionary[input.name] = input
        else:
            raise TypeError(f"Unexpected input type {type(input)} for add_operator.")

    def apply_operator(self, prefix_sentence):
        if len(prefix_sentence) == 1:
            atom = prefix_sentence[0]
            if atom in {"\\top", "\\bot"}:  # Handle extremal operators
                return [self[atom]]
            if atom.isalnum():  # Handle atomic sentences
                return [Const(atom, AtomSort)]
            raise ValueError(f"The atom {atom} is invalid in apply_operator.")
        op, arguments = prefix_sentence[0], prefix_sentence[1:]
        activated = [self.apply_operator(arg) for arg in arguments]
        if isinstance(op, str):
            activated.insert(0, self[op])
        else:
            raise TypeError(f"Expected operator name as a string, got {type(op).__name__}.")
        return activated

    # # NOTE: UPDATE OP STRATEGY
    # def duplicate(self):
    #     """Creates a shallow copy of the OperatorCollection."""
    #     new_collection = OperatorCollection()
    #     new_collection.operator_dictionary = self.operator_dictionary.copy()
    #     return new_collection
    #
    # # NOTE: UPDATE OP STRATEGY
    # def update_operators(self, semantics):
    #     operators = self.operator_dictionary
    #     for key in operators.keys():
    #         if isinstance(operators[key], type):
    #             operators[key] = operators[key](semantics)


class Syntax:
    """Takes infix_premises, infix_conclusions, and operator_collection as
    arguments, generating and storing instances of the Sentence class.
    Draws on Sentence instances for the premises and conclusions in order to
    store a dictionary which includes all subsentences for each before drawing
    on the operator_collection in order to initialize the original_types in each
    sentence object in the dictionary. Lists are then stored for the premises,
    conclusions, and all sentence_letters that occur in theses sentences."""

    def __init__(
        self,
        infix_premises,
        infix_conclusions,
        operator_collection,
    ):

        # start timer
        self.start_time = time.time()

        # store inputs
        self.infix_premises = infix_premises
        self.infix_conclusions = infix_conclusions
        self.operator_collection = operator_collection

        # initialize inputs
        self.all_sentences = {} # updated in build_sentence
        # self.operators_used = []
        self.sentence_letters = [] # updated in build_sentence
        self.premises = self.initialize_sentences(self.infix_premises)
        self.conclusions = self.initialize_sentences(self.infix_conclusions)

        # check for interdefined operators
        self.circularity_check(operator_collection)

    def initialize_sentences(self, infix_sentences):
        """Takes a list of sentences composing the dictionaries of subsentences
        for each, resulting in a dictionary that includes all subsentences."""

        def build_sentence(infix_sentence):
            if infix_sentence in self.all_sentences.keys():
                return self.all_sentences[infix_sentence]
            sentence = Sentence(infix_sentence)
            # TODO: confirm not needed
            # if sentence.original_operator:
            #     self.operators_used.append(sentence.original_operator)
            self.all_sentences[sentence.name] = sentence
            if sentence.original_arguments is None:
                if sentence.name.isalnum():
                    self.sentence_letters.append(sentence)
                return sentence
            sentence_arguments = []
            for infix_arg in sentence.original_arguments:
                sentence_arg = build_sentence(infix_arg)
                sentence_arguments.append(sentence_arg)
            sentence.original_arguments = sentence_arguments
            return sentence

        def initialize_types(sentence):
            """Draws on the operator_collection in order to initialize all sentences
            in the input by replacing operator strings with operator classes and
            updating original_type in that sentence_obj. If the main operator is not
            primitive, derived_arguments are updated with derived_types."""
            # TODO: confirm not needed with derived operators
            # if sentence.original_arguments:
            #     for argument in sentence.original_arguments:
            #         initialize_types(argument)
            sentence.update_types(self.operator_collection)
            if sentence.arguments: # NOTE: must happen after arguments are stored
                sentence_arguments = []
                for infix_arg in sentence.arguments:
                    argument = build_sentence(infix_arg)
                    initialize_types(argument)
                    sentence_arguments.append(argument)
                sentence.arguments = sentence_arguments 

        sentence_objects = []
        for infix_sent in infix_sentences:
            # TODO: this check/continue leads to errors
            # if infix_sent in self.all_sentences.keys():
            #     continue
            sentence = build_sentence(infix_sent)
            initialize_types(sentence)
            sentence_objects.append(sentence)
        return sentence_objects

    def circularity_check(self, operator_collection):
        """Detects circular dependencies among operators and ensures all dependencies are within the collection."""
        dependency_graph = {}
        operator_set = set(operator_collection.operator_dictionary.values())

        # Build the dependency graph
        for operator_class in operator_collection.operator_dictionary.values():
            if operator_class.primitive:
                continue
            # Ensure derived_definition is callable
            if isinstance(operator_class, type) and callable(getattr(operator_class, 'derived_definition', None)):
                sig = inspect.signature(operator_class.derived_definition)
                num_params = len(sig.parameters)
                dummy_args = [None] * num_params
                dependencies = {
                    elem for elem in flatten(operator_class.derived_definition(*dummy_args))
                    if isinstance(elem, type)
                }
                # Identify missing dependencies if any
                missing_dependencies = dependencies - operator_set
                if missing_dependencies:
                    missing_names = [op.name for op in missing_dependencies]
                    raise ValueError(
                        f"Operator '{operator_class.name}' depends on undefined operators: {missing_names}"
                    )
                # Update the dependency graph
                if operator_class not in dependency_graph:
                    dependency_graph[operator_class] = set()
                dependency_graph[operator_class].update(dependencies)
                continue
            raise ValueError(
                f"Error processing operator '{operator_class.name}'"
            )

        visited = set()
        recursion_stack = []

        def dfs(current):
            if current in recursion_stack:
                cycle = " -> ".join(op.name for op in recursion_stack) + f" -> {current.name}"
                raise RecursionError(
                    f"Circular definition detected in {current.name}.\n\nCYCLE: {cycle}\n"
                )
            if current in visited:
                return
            recursion_stack.append(current)
            for dependent in dependency_graph.get(current, set()):
                dfs(dependent)
            recursion_stack.remove(current)
            visited.add(current)

        # Perform DFS for each operator to detect cycles
        for operator in operator_collection.operator_dictionary.values():
            if not operator.primitive and operator not in visited:
                dfs(operator)
